function T = horn(L, R)
% HORN Horn's method for absolute orientation.
%   HORN(L, R) solves the 3D absolute orientation problem for
%   the data sets L and R where L and R are 3xN arrays of
%   column vectors. Returns the rigid homogeneous transformation
%   matrix T mapping L to R.

% number of points
n = size(L, 2);

% centroids of L and R
Lbar = mean(L, 2);
Rbar = mean(R, 2);

% subtract centroids from L and R
L = L - repmat(Lbar, 1, n);
R = R - repmat(Rbar, 1, n);

% compute sums of products
M = zeros(3, 3);
for a = 1:n
  M = M + L(:, a) * R(:, a).';
end

% elements of M labelled for convenience
sxx = M(1, 1);
sxy = M(1, 2);
sxz = M(1, 3);
syx = M(2, 1);
syy = M(2, 2);
syz = M(2, 3);
szx = M(3, 1);
szy = M(3, 2);
szz = M(3, 3);

% compute N
N = zeros(4, 4);
N(1, 1) = sxx + syy + szz;
N(1, 2) = syz - szy;
N(1, 3) = szx - sxz;
N(1, 4) = sxy - syx;
N(2, 2) = sxx - syy - szz;
N(2, 3) = sxy + syx;
N(2, 4) = szx + sxz;
N(3, 3) = -sxx + syy - szz;
N(3, 4) = syz + szy;
N(4, 4) = -sxx - syy + szz;
N(2, 1) = N(1, 2);
N(3, 1) = N(1, 3);
N(3, 2) = N(2, 3);
N(4, 1) = N(1, 4);
N(4, 2) = N(2, 4);
N(4, 3) = N(3, 4);

% find the eigenvector corresponding to the largest eigenvalue
[V, D] = eig(N);
d = [D(1,1) D(2,2) D(3,3) D(4,4)];
[y, idx] = max(d);
q = V(:, idx)';
if(q(1) < 0)
  q = -q;
end

% compute residuals
ROT = mq(q);
t = Rbar - (ROT * Lbar);
T = [ROT t; 0 0 0 1];


